import time
import serial
from datetime import datetime
from datetime import timedelta
import os

class ModemDriver:
    serialDevice: serial.Serial
    listenerLog = []
    senderLog = []
    waitTime = 8.0
    
    @property
    def device_speed(self):
        return self._speed

    @property
    def device_name(self):
        return self._device
        
    def __init__(self, com_port: str, sendTone:bool=True):
        self._sending_tone = False

        if sendTone:
            self._dial_tone_wav = self._read_dial_tone()
        else:
            self._dial_tone_wav = None
        #110, 300, 1200, 2400, 4800, 9600, 19200, 38400, 57600, or 115200
        speed =  115200 #209954 

        self._device, self._speed = com_port, speed
  
  
        self.serialDevice = None
        print("Modem:Opening netlink serial interface to {}".format(com_port))
        
        
        self.serialDevice = serial.Serial(com_port, 57600, timeout=0.01, rtscts = True,write_timeout =1)
        
        #print("potential modem found at %s" % com)
        self.query_modem("AT",timeout = 1) # potential modem. Other devices respond to AT, so not definitive.
        self.query_modem("AT+FCLASS=8",timeout = 1) # if potential modem, find out if it's our voice modem
        self.reset()
       
       
        
        
    def query_modem(self, command, timeout=3, response = "OK"): #this function assumes we're being passed a non-blocking modem
        if isinstance(command, bytes):
            final_command = command + b'\r\n'
        else:
            final_command = ("%s\r\n" % command).encode()
            final_command = ("%s\r\n" % command).encode()
        print(final_command.decode())
            
        self.serialDevice.write(final_command)
        

        start = time.time()

        line = b""
        while True:
            new_data = self.serialDevice.readline().strip()

            if not new_data: #non-blocking modem will end up here when timeout reached, try until this function's timeout is reached.
                if time.time() - start < timeout:
                    continue
                print("Raising IO Error")
                raise IOError()

            line = line + new_data
            print("Response: %s" % line)
            if response.encode() in line:
                if response != "OK":
                    print(line.decode())
                return  # Valid response

    def reset(self):
        self.send_command("ATZ0")  # Send reset command
        time.sleep(1)
        self.send_command("AT&F0")
        self.send_command("ATE0W2")  # Don't echo our responses
    
    def send_command(self, command, timeout=60, ignore_responses=None):
        ignore_responses = ignore_responses or []  # Things to completely ignore

        VALID_RESPONSES = [b"OK", b"ERROR", b"CONNECT", b"VCON"]

        for ignore in ignore_responses:
            VALID_RESPONSES.remove(ignore.encode())

        final_command = ("%s\r\n" % command).encode()
        self.serialDevice.write(final_command)
        print(final_command.decode())

        start = datetime.now()

        line = b""
        while True:
            new_data = self.serialDevice.readline().strip() #this is blocking

            if not new_data:
                continue

            line = line + new_data
            for resp in VALID_RESPONSES:
                if resp in line:
                    if resp != b"OK":
                        print(line.decode())
                    # logger.info(line[line.find(resp):])
                    return  # We are done

            if (datetime.now() - start).total_seconds() > timeout: #if readline gets hung up, this if will never be reached
                raise IOError("There was a timeout while waiting for a response from the modem")

   
    
    def read(self) -> bytearray:
        waiting = self.serialDevice.in_waiting
        data = self.serialDevice.read(waiting)
        if not data:
            data = b""
        else:
            self.senderLog.append(data)
        return data     

        
    def write(self,data):
        self.serialDevice.write(data)
        self.listenerLog.append(data)
        

        
    def send_escape(self):
        time.sleep(1.0)
        self.serialDevice.write(b"+++")
        time.sleep(1.0)
        
    def connect(self):
        if self.serialDevice:
            self.disconnect()

        print("Opening serial interface to {}".format(self._device))
        self.serialDevice = serial.Serial(
            self._device, self._speed, timeout=0
        )
        self.send_escape()
        self.start_dial_tone()
       
        print("Flushing")
        self.serialDevice.flush()
        self.serialDevice.reset_input_buffer()
        self.serialDevice.reset_output_buffer()
        time.sleep(1)
        
        while True:
            char = self.read()
            if not char:
                break
        
        print("Waiting for device response")
        #self.send_command("AT0")
        while True:
            self.update()
            
            char = self.read()
            
            if not char:
                continue
            print("%s" % char)
            
            if b'\x10' in char:
                # DLE character
                try:
                    if not b'*' in char:
                        char = self.read()
                        
                    if not char or not b'*' in char:
                        continue
                    
                    print("stopping dial tone")
                    self.stop_dial_tone()
                    time_digit_heard = datetime.now()
                    while (datetime.now() - time_digit_heard).total_seconds() <= self.waitTime:
                        continue;
                        
                    print("Done waiting")
                    
                    time_digit_heard = None

                    if self.serialDevice:
                        self.disconnect()
                
                    print("Opening netlink serial interface to {}".format(self._device))
                    self.serialDevice = serial.Serial(self._device, 57600, timeout=0.01, rtscts=True)
                
                
                    self.query_modem(b'AT%E0')
                    self.query_modem(b"AT\N3\V1%C0")
                    self.query_modem(b'AT+MS=V32b,1,14400,14400,14400,14400')
                    self.query_modem(b"ATA", timeout=120, response = "CONNECT")
                    
                    print("Finished connecting")
                    return True
                except IOError as error:
                    print(IOError)
                    return False
                    
                except (TypeError, ValueError):
                    print(TypeError)
                    print(ValueError)
                    pass    

    def disconnect(self):
        try:     
            
            self.serialDevice.flush()
            for i in range(3):
                self.write(b'+')
                time.sleep(0.2)
            time.sleep(4)
             
        
            if self.serialDevice.isOpen():
                self.serialDevice.close()

            self.serialDevice = None
            f = open("sender%s.txt" % self._device, "wb")
            for line in self.senderLog:
                f.write(line)
            f.close()
            
            f = open("listener%s.txt" % self._device, "wb")
            for line in self.listenerLog:
                f.write(line)
            f.close()
            
        except IOError as error:
            pass
        finally:
            pass
        
    def update(self):
        now = datetime.now()
        if self._sending_tone:
            # Keep sending dial tone
            BUFFER_LENGTH = 1000
            TIME_BETWEEN_UPLOADS_MS = (1000.0 / 8000.0) * BUFFER_LENGTH

            milliseconds = (now - self._time_since_last_dial_tone).microseconds * 1000
            if not self._time_since_last_dial_tone or milliseconds >= TIME_BETWEEN_UPLOADS_MS:
                byte = self._dial_tone_wav[self._dial_tone_counter:self._dial_tone_counter+BUFFER_LENGTH]
                self._dial_tone_counter += BUFFER_LENGTH
                if self._dial_tone_counter >= len(self._dial_tone_wav):
                    self._dial_tone_counter = 0
                self.serialDevice.write(byte)
                self._time_since_last_dial_tone = now
     
    def start_dial_tone(self):
        if not self._dial_tone_wav:
            return

        self.reset()
        self.send_command("AT+FCLASS=8")  # Enter voice mode
        self.send_command("AT+VLS=1")  # Go off-hook
        self.send_command("AT+VSM=1,8000")  # 8 bit unsigned PCM
        self.send_command("AT+VTX")  # Voice transmission mode

        self._sending_tone = True

        self._time_since_last_dial_tone = (
            datetime.now() - timedelta(seconds=100)
        )

        self._dial_tone_counter = 0

    def stop_dial_tone(self):
        if not self._sending_tone:
            return

        self.serialDevice.write(("\0{}{}\r\n".format(chr(0x10), chr(0x03))).encode())
        self.send_escape()
        self.send_command("ATH0")  # Go on-hook
        self.reset()  # Reset the modem
        self._sending_tone = False
    def _read_dial_tone(self):
        this_dir = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))
        dial_tone_wav = os.path.join(this_dir, "dial-tone.wav")

        with open(dial_tone_wav, "rb") as f:
            dial_tone = f.read()  # Read the entire wav file
            dial_tone = dial_tone[44:]  # Strip the header (44 bytes)

        return dial_tone
        
    def flush(self):
        self.serialDevice.flush()